"""
Coded By Mohammad Taha Gorji
SafeRun: A highly secure sandbox for executing Python code in an isolated environment.

This library can be imported in your projects:
    from SafeRun import UltraSecureSandbox

It can also be executed as a module:
    python -m SafeRun <file.py> [--time TIME] [--memory MEMORY] [--chroot CHROOT_DIR]

Note:
    - For full functionality (e.g., chroot and privilege dropping), this module must be run with appropriate permissions.
    - The chroot directory must be pre-configured and secured.
    - This implementation uses multiple techniques (AST checks, resource limits, chroot, privilege dropping, and optional seccomp filters)
      to maximize security. However, due to the dynamic nature of Python, no sandbox can be 100% secure.
"""

import ast
import os
import resource
import multiprocessing
import traceback

# Attempt to import seccomp for syscall filtering (optional)
try:
    import seccomp
    HAS_SECCOMP = True
except ImportError:
    HAS_SECCOMP = False


class SandboxException(Exception):
    """Exception for errors in the SafeRun sandbox."""
    pass


def check_ast(node):
    """
    Perform a comprehensive AST check on the provided code node.
    
    Disallowed:
      - Any import or import-from statements.
      - Calls to dangerous functions: eval, exec, __import__, open, compile, input, globals, locals, vars, dir.
      - Access to attributes that start and end with '__' (e.g., __globals__).
      - Lambda expressions.
    """
    dangerous_calls = {'eval', 'exec', '__import__', 'open', 'compile', 'input', 'globals', 'locals', 'vars', 'dir'}
    for child in ast.walk(node):
        if isinstance(child, (ast.Import, ast.ImportFrom)):
            raise SandboxException("Import statements are not allowed.")
        if isinstance(child, ast.Call):
            if isinstance(child.func, ast.Name) and child.func.id in dangerous_calls:
                raise SandboxException(f"Call to {child.func.id} is not allowed.")
        if isinstance(child, ast.Attribute):
            attr = child.attr
            if attr.startswith("__") and attr.endswith("__"):
                raise SandboxException(f"Access to {attr} is not allowed.")
        if isinstance(child, ast.Lambda):
            raise SandboxException("Lambda expressions are not allowed.")
    return True


def drop_privileges(uid=65534, gid=65534):
    """
    Drop privileges by setting the UID and GID.
    
    Default values (65534) usually correspond to the 'nobody' user on Unix systems.
    """
    os.setgid(gid)
    os.setuid(uid)


def setup_seccomp_filter():
    """
    Set up a seccomp filter if the seccomp library is available.
    
    Only a minimal set of syscalls necessary for execution are allowed.
    """
    if not HAS_SECCOMP:
        return
    filt = seccomp.SyscallFilter(defaction=seccomp.SCMP_ACT_KILL)
    safe_syscalls = [
        "read", "write", "exit", "exit_group", "sigreturn",
        "rt_sigreturn", "rt_sigprocmask", "brk", "mmap", "munmap",
        "close", "futex", "nanosleep", "getpid"
    ]
    for syscall in safe_syscalls:
        try:
            filt.add_rule(seccomp.SCMP_ACT_ALLOW, syscall)
        except Exception:
            pass
    filt.load()


def run_in_sandbox(code, safe_globals, time_limit, memory_limit, chroot_dir):
    """
    Execute the given code in a sandboxed environment using the following techniques:
      - Resource limits (CPU time and memory) via the resource module.
      - Filesystem isolation using chroot.
      - Privilege dropping to a non-privileged user.
      - Optional syscall filtering using seccomp.
    
    Returns:
      The output message or traceback from the sandboxed execution.
    """
    def target(queue):
        try:
            resource.setrlimit(resource.RLIMIT_CPU, (time_limit, time_limit))
            resource.setrlimit(resource.RLIMIT_AS, (memory_limit, memory_limit))
            os.chroot(chroot_dir)
            os.chdir("/")
            drop_privileges()
            setup_seccomp_filter()
            compiled_code = compile(code, "<sandbox>", "exec")
            exec(compiled_code, safe_globals)
            queue.put("Execution finished successfully.")
        except Exception:
            queue.put(traceback.format_exc())

    queue = multiprocessing.Queue()
    p = multiprocessing.Process(target=target, args=(queue,))
    p.start()
    p.join(time_limit + 2)
    if p.is_alive():
        p.terminate()
        return "Error: Timeout reached."
    if not queue.empty():
        return queue.get()
    return "Error: No response from sandbox."


class UltraSecureSandbox:
    """
    UltraSecureSandbox provides a highly secure environment for executing Python code.

    Features:
      - Comprehensive AST checking to prevent dangerous code patterns.
      - CPU and memory usage restrictions.
      - Execution in a separate process.
      - Filesystem isolation using chroot.
      - Privilege dropping.
      - Optional seccomp syscall filtering.
      
    Usage:
        sandbox = UltraSecureSandbox(time_limit=2, memory_limit_mb=50, chroot_dir="/tmp/sandbox_chroot")
        sandbox.execute_file("example.py")
    """
    def __init__(self, allowed_builtins=None, time_limit=2, memory_limit_mb=50, chroot_dir="/tmp/sandbox_chroot"):
        if allowed_builtins is None:
            allowed_builtins = {
                'print': print,
                'abs': abs,
                'min': min,
                'max': max,
                'sum': sum,
                'range': range,
                'len': len,
            }
        self.safe_globals = {"__builtins__": allowed_builtins}
        self.time_limit = time_limit
        self.memory_limit = memory_limit_mb * 1024 * 1024  # Convert MB to bytes
        self.chroot_dir = chroot_dir

    def execute_code(self, code):
        try:
            tree = ast.parse(code, mode='exec')
            check_ast(tree)
        except SandboxException as se:
            print(f"Sandbox AST error: {se}")
            return
        except Exception as e:
            print(f"AST parse error: {e}")
            return

        result = run_in_sandbox(code, self.safe_globals, self.time_limit, self.memory_limit, self.chroot_dir)
        print(result)

    def execute_file(self, file_path):
        try:
            with open(file_path, "r", encoding="utf-8") as f:
                code = f.read()
            self.execute_code(code)
        except FileNotFoundError:
            print("File not found.")
        except Exception as e:
            print(f"File read error: {e}")


def main():
    import argparse
    parser = argparse.ArgumentParser(
        description="SafeRun: Execute Python code securely in an isolated environment."
    )
    parser.add_argument("file", help="Path to the Python file to execute securely.")
    parser.add_argument("--time", type=int, default=2, help="CPU time limit in seconds.")
    parser.add_argument("--memory", type=int, default=50, help="Memory limit in MB.")
    parser.add_argument("--chroot", type=str, default="/tmp/sandbox_chroot",
                        help="Chroot directory (must be pre-configured and secured).")
    args = parser.parse_args()
    sandbox = UltraSecureSandbox(time_limit=args.time, memory_limit_mb=args.memory, chroot_dir=args.chroot)
    sandbox.execute_file(args.file)


# Allow both import and command-line execution.
if __name__ == "__main__":
    main()
